#version 330 core

#extension GL_NV_shadow_samplers_cube : enable

in vec2 vert_texCoords;
in vec3 vert_normal;
in vec3 vert_fragPos;
in mat3 vert_tbn;
in vec4 viewSpace;

out vec4 out_fragColor;

uniform sampler2D gPosition;
uniform sampler2D gNormal;
uniform sampler2D texNoise;
uniform sampler2D ssao;


// Texturing
uniform bool u_useLight;
uniform vec3 u_color = vec3(1.0f, 0.0f, 1.0f); // magenta

uniform sampler2D u_texDiffuse; // 0
uniform bool u_useDiff;

uniform sampler2D u_texSpecular; // 1
uniform bool u_useSpec;

uniform sampler2D u_texNormal; // 2
uniform bool u_useNormal;

uniform samplerCube u_texShadowsCubemaps[10]; // 3, 4, 5, 6, 7, 8, 9, 10, 11, 12
uniform bool u_doShadowMappingPoint;
uniform bool u_shadowsLightActive[10];
uniform float u_shadowsFar;
const float SHADOWS_BIAS = 0.05f;

// Fog

//0 linear; 1 exponential; 2 exponential square
uniform int fogSelector;
//0 plane based; 1 range based
uniform int depthFog;
uniform vec3 u_cameraPos;

const vec3 fogColor = vec3(0.4f,0.4f,0.4f);
const float fogDensity = 0.03f;

// Lighting
struct PointLight {
	vec3 pos;
	vec3 col;
	float spec;
	// attenuation
	float cons;
	float lin;
	float quad;
	//bool castsShadows;
};
uniform PointLight u_pointLights[10];
uniform int u_pointLightsCount = 0;

uniform vec3 u_viewPos; // camera pos
uniform float u_lightAmb;  // ambient lighting intensity

// Bloom
uniform float u_bloomThresh = 0.8f;
const vec3 WEIGHTS = vec3(0.2126f, 0.7152f, 0.0722f); // RGB weights for lum

// Member variables
vec3 normal;
vec3 color;

vec3 illuminate(PointLight light) {
	vec3 fragToLight = light.pos - vert_fragPos;
	vec3 lightDir = normalize(fragToLight);
	float dis = length(fragToLight);
	
	float diff = max(dot(normal, lightDir), 0.0f); // clamp to zero
	vec3 light_diffuse = diff * light.col;
	
	vec3 light_specular;
	if (light.spec > 0.0f) {
		vec3 viewDir = normalize(u_viewPos - vert_fragPos);
		
		/* // PHONG
		vec3 reflDir = reflect(-lightDir, normal); // reflection direction
		// 32 here represents the shininess: the size of the specular circle
		// Check an image here: https://learnopengl.com/Lighting/Basic-Lighting
		float spec = pow(max(dot(viewDir, reflDir), 0.0f), 32);
		*/
		
		// BLINN-PHONG
		vec3 halfwayDir = normalize(lightDir + viewDir);
		float spec = pow(max(dot(normal, halfwayDir), 0.0f), 64.0f);
		
		light_specular = light.spec * spec * light.col;
		
	} else {
		light_specular = vec3(0.0f, 0.0f, 0.0f);
	}
	if (u_useSpec) {
		light_specular *= texture2D(u_texSpecular, vert_texCoords).r;
	}
	
	float attenuation = 1.0f / (light.cons + light.lin * dis + light.quad * (dis * dis));
	
	return attenuation * (light_diffuse + light_specular);
}

void main() {

	if (u_useNormal) {
		normal = texture2D(u_texNormal, vert_texCoords).xyz;
		normal = normalize(normal * 2.0f - 1.0f);
		normal = normalize(vert_tbn * normal);
	} else {
		normal = normalize(vert_normal);
	}
	
	color = u_useDiff ? (texture2D(u_texDiffuse, vert_texCoords).rgb) : u_color;
	
	// Assuming u_texDiffuse has at least 3 components
	vec3 colorFinal = vec3(0.0f, 0.0f, 0.0f);
	for (int i = 0; i < u_pointLightsCount; i++) {
		float shadow = 0.0f;
		if (u_doShadowMappingPoint && u_shadowsLightActive[i]) {
			vec3 lightToFrag = vert_fragPos - u_pointLights[i].pos;
			float closestDepth = textureCube(u_texShadowsCubemaps[i], lightToFrag).r * u_shadowsFar;
			float currentDepth = length(lightToFrag);
			shadow = (currentDepth - SHADOWS_BIAS) > closestDepth ? 1.0f : 0.0f;
		}
		colorFinal += color * (u_lightAmb + (1.0f - shadow) * illuminate(u_pointLights[i]));
	}

	//out_fragColor = vec4(colorFinal, 1.0f);
	
	vec3 finalColor = vec3(0.0f, 0.0f, 0.0f);
	
	//Fog
	float dist = 0.0f;
	float fogFactor = 0.0f;
	
	// float be = (u_cameraPos.y - viewSpace.y)*0.004;
	// float bi = (u_cameraPos.y - viewSpace.y)*0.001;
	
	float be = 15.0f*0.004;
	float bi = 15.0f*0.001;
	
	//compute distance used in fog equations
	if(depthFog == 0){
		//plane based
		dist = abs(viewSpace.z);
	}else{
		//range based
		dist = length(viewSpace);
	}
	
	//type of fog equation
	if(fogSelector == 0){ //linear fog 
		// 4 - fog starts; 15 - fog ends
		fogFactor = (15.0f - dist)/(15.0f - 4.0f);
		fogFactor = clamp(fogFactor, 0.0f, 1.0f);
		
		//if you inverse color in glsl mix function you have to put 1.0 - fogFactor
		finalColor = mix(fogColor, colorFinal, fogFactor);
	}else if(fogSelector == 1){ //exponential fog 
		fogFactor = 1.0f/exp(dist*fogDensity);
		fogFactor = clamp(fogFactor, 0.0f, 1.0f);
		
		if(vert_fragPos.z < -30.0f){
			fogFactor = 1;
		}
		
		// if(vert_fragpos.y > 0.0f){
			// fogFactor = 1;
		// }
		// if(vert_fragpos.y > -0.5f){
			// fogFactor = 0.9;
		// }
		// if(vert_fragpos.y > -1.0f){
			// fogFactor = 0.8;
		// }
		// if(vert_fragpos.y > -1.5f){
			// fogFactor = 0.7;
		// }
		// if(vert_fragpos.y > -2.0f){
			// fogFactor = 0.6;
		// }

		//mix function fogColor*(1-fogFactor) + lightcolor*fogFactor;
		finalColor = mix(fogColor, colorFinal, fogFactor);
	}else if(fogSelector == 2){ // exponential square fog 
		fogFactor = 1.0f/exp((dist*fogDensity)*(dist*fogDensity));
		fogFactor = clamp(fogFactor, 0.0f, 1.0f);
		
		finalColor = mix(fogColor, colorFinal, fogFactor);
	}
	
	float ext = exp(-dist*be);
	float insc = exp(-dist*bi);
	
	finalColor = colorFinal*ext + fogColor*(1-insc);
	
	out_fragColor = vec4(finalColor, 1.0f);
		
}