#version 330 core

#extension GL_NV_shadow_samplers_cube : enable

in vec2 vert_texCoords;

out vec4 out_fragColor;

uniform sampler2D u_texColorSpec; // 0
uniform sampler2D u_texFragPos; // 1
uniform sampler2D u_texNormal; // 2
uniform sampler2D u_texViewSpace; // 3
uniform sampler2D u_texAmbientOcclusion; // 4

// Lighting
struct PointLight {
	vec3 pos;
	vec3 col;
	float spec;
	// attenuation
	float cons;
	float lin;
	float quad;
	//bool castsShadows;
};
uniform PointLight u_pointLights[10];
uniform int u_pointLightsCount = 0;
uniform float u_lightAmb;  // ambient lighting intensity

uniform vec3 u_camPos; // camera pos

// Shadow mapping
uniform samplerCube u_texShadowsCubemaps[10]; // 5-14
uniform bool u_doShadowMappingPoint;
uniform bool u_shadowsLightActive[10];
uniform float u_shadowsFar[10];
const float SHADOWS_BIAS = 0.15f;

// Fog
uniform bool u_doFog;
//0 linear; 1 exponential; 2 exponential square
uniform int u_fogSelector;
//0 plane based; 1 range based
uniform int u_depthFog;

const vec3 u_fogColor = vec3(0.5f,0.5f,0.5f);
const float u_fogDensity = 0.06f;

// Ambient occlusion
uniform bool u_doAmbientOcclusion;

// Member variables --------------------------------------------------------------------------------------
vec3 texColor;
float texSpec;
vec3 texFragPos;
vec3 texNormal;
vec4 texViewSpace;
float texAmbientOcclusion;

// Illuminate --------------------------------------------------------------------------------------------
vec3 illuminate(PointLight light) {
	vec3 fragToLight = light.pos - texFragPos;
	vec3 lightDir = normalize(fragToLight);
	float dis = length(fragToLight);
	
	float diff = max(dot(texNormal, lightDir), 0.0f); // clamp to zero
	vec3 light_diffuse = diff * light.col;
	
	vec3 light_specular;
	if (light.spec > 0.0f) {
		vec3 viewDir = normalize(u_camPos - texFragPos);
		
		/* // PHONG
		vec3 reflDir = reflect(-lightDir, normal); // reflection direction
		// 32 here represents the shininess: the size of the specular circle
		// Check an image here: https://learnopengl.com/Lighting/Basic-Lighting
		float spec = pow(max(dot(viewDir, reflDir), 0.0f), 32);
		*/
		
		// BLINN-PHONG
		vec3 halfwayDir = normalize(lightDir + viewDir);
		float spec = pow(max(dot(texNormal, halfwayDir), 0.0f), 64.0f);
		
		light_specular = light.spec * spec * light.col;
		
	} else {
		light_specular = vec3(0.0f, 0.0f, 0.0f);
	}
	light_specular *= texSpec;
	
	float attenuation = 1.0f / (light.cons + light.lin * dis + light.quad * (dis*dis));
	
	return attenuation * (light_diffuse + light_specular);
}

// Fog ---------------------------------------------------------------------------------------------------
vec3 applyFog(vec3 color) {
	vec3 finalColor = vec3(0.0f, 0.0f, 0.0f);
	float dist = 0.0f;
	float fogFactor = 0.0f;
	
	// float be = (u_cameraPos.y - viewSpace.y)*0.004;
	// float bi = (u_cameraPos.y - viewSpace.y)*0.001;
	
	float be = 15.0f*0.004;
	float bi = 15.0f*0.001;
	
	//compute distance used in fog equations
	if(u_depthFog == 0){
		//plane based
		dist = abs(texViewSpace.z);
	}else{
		//range based
		dist = length(texViewSpace);
	}
	
	//type of fog equation
	switch (u_fogSelector) {
	case 0: //linear fog 
		// 4 - fog starts; 15 - fog ends
		fogFactor = (15.0f - dist)/(15.0f - 1.0f);
		fogFactor = clamp(fogFactor, 0.0f, 1.0f);
		
		//if you inverse color in glsl mix function you have to put 1.0 - fogFactor
		finalColor = mix(u_fogColor, color, fogFactor);
		break;
		
	case 1: //exponential fog 
		fogFactor = 1.0f/exp(dist*u_fogDensity);
		fogFactor = clamp(fogFactor, 0.0f, 1.0f);
		
		if (texFragPos.z < -30.0f){
			fogFactor = 1;
		}
		
		// if(vert_fragpos.y > 0.0f){
			// fogFactor = 1;
		// }
		// if(vert_fragpos.y > -0.5f){
			// fogFactor = 0.9;
		// }
		// if(vert_fragpos.y > -1.0f){
			// fogFactor = 0.8;
		// }
		// if(vert_fragpos.y > -1.5f){
			// fogFactor = 0.7;
		// }
		// if(vert_fragpos.y > -2.0f){
			// fogFactor = 0.6;
		// }

		//mix function fogColor*(1-fogFactor) + lightcolor*fogFactor;
		finalColor = mix(u_fogColor, color, fogFactor);
		break;
		
	case 2: // exponential square fog 
		fogFactor = 1.0f/exp((dist*u_fogDensity)*(dist*u_fogDensity));
		fogFactor = clamp(fogFactor, 0.0f, 1.0f);
		
		finalColor = mix(u_fogColor, color, fogFactor);
		break;
	
	case 3:
		float ext = exp(-dist*be);
		float insc = exp(-dist*bi);
		if(texFragPos.z > -60.0f){
			finalColor = color*ext + u_fogColor*(1-insc);
		}else{
			finalColor = color;
		}
		break;
		
	default:
		finalColor = color;
		break;
	}
	
	return finalColor;
}

// Main --------------------------------------------------------------------------------------------------
void main() {

		
	vec4 texColorSpec = texture2D(u_texColorSpec, vert_texCoords);
	texColor = texColorSpec.rgb;
	texSpec = texColorSpec.w;
	texFragPos = texture2D(u_texFragPos, vert_texCoords).rgb;
	texNormal = texture2D(u_texNormal, vert_texCoords).rgb;
	texViewSpace = texture2D(u_texViewSpace, vert_texCoords);
	texAmbientOcclusion = texture2D(u_texAmbientOcclusion, vert_texCoords).r;
	
	vec3 color = vec3(0.0f, 0.0f, 0.0f);
	for (int i = 0; i < u_pointLightsCount; i++) {
		float shadow = 0.0f;
		if (u_doShadowMappingPoint && u_shadowsLightActive[i]) {
			vec3 lightToFrag = texFragPos - u_pointLights[i].pos;
			float closestDepth = textureCube(u_texShadowsCubemaps[i], lightToFrag).r * u_shadowsFar[i];
			float currentDepth = length(lightToFrag);
			shadow = (currentDepth - SHADOWS_BIAS) > closestDepth ? 1.0f : 0.0f;
		}
		color += texColor * (u_lightAmb + (1.0f - shadow) * illuminate(u_pointLights[i]));
	}
	
	// ambient occlusion
	if (u_doAmbientOcclusion) {
		color *= texAmbientOcclusion;
	}
	
	// Fog
	if (u_doFog) {
		color = applyFog(color);
	}
	
	out_fragColor = vec4(color, 1.0f);
}